"""
Two basic data types are defined for forward propagation and back propagation.
Signed and unsigned fixed point data needed Straight-Through filter for back propagation.
"""

import torch
from torch.autograd import Function
import math
from ._utils import *


class Ceil(Function):
    @staticmethod
    def forward(ctx, input):
        output = torch.ceil(input)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        return torch.ones(grad_output.shape).type_as(grad_output)


ceil = Ceil.apply


class RoundToEven(Function):
    @staticmethod
    def forward(ctx, input):
        round = torch.floor(input)
        return round + (round % 2)

    @staticmethod
    def backward(ctx, grad_output):
        return torch.ones(grad_output.shape).type_as(grad_output)


round = RoundToEven.apply


class qSigned(Function):
    @staticmethod
    def forward(ctx, x, log2_t, bit_width):
        r'''
        See TQT paper's Eqn. (4)
        '''
        bit_max = 2.**(bit_width - 1)
        n = -bit_max
        p = bit_max - 1
        s = 2.**ceil(log2_t) / bit_max
        q = torch.clamp(round(x / s), n, p) * s
        ctx.save_for_backward(x / s, s, number_to_tensor(n, x),
                              number_to_tensor(p, x))
        return q

    @staticmethod
    def backward(ctx, grad_output):
        x_div_s, s, n, p = ctx.saved_tensors
        rounded = round(x_div_s)
        cmp0 = (n <= rounded) & (rounded <= p)
        cmp1 = rounded < n
        cmp2 = rounded > p
        grad_s = (rounded - x_div_s) * cmp0 + n * cmp1 + p * cmp2
        grad_log_2_t = math.log(2) * grad_s
        grad_x = cmp0 * 1.0
        return grad_output * grad_x, grad_output * grad_log_2_t, None


qsigned = qSigned.apply


class qUnsigned(Function):
    @staticmethod
    def forward(ctx, x, log2_t, bit_width):
        r'''
        See TQT paper's Eqn. (4)
        '''
        bit_max = 2.**(bit_width)
        n = 0
        p = bit_max - 1
        s = 2.**ceil(log2_t) / bit_max
        q = torch.clamp(round(x / s), n, p) * s
        ctx.save_for_backward(round(x / s), s, number_to_tensor(n, x),
                              number_to_tensor(p, x))
        return q

    @staticmethod
    def backward(ctx, grad_output):
        x_div_s, s, n, p = ctx.saved_tensors
        rounded = round(x_div_s)
        cmp0 = (n <= rounded) & (rounded <= p)
        cmp1 = rounded < n
        cmp2 = rounded > p
        grad_s = (rounded - x_div_s) * cmp0 + n * cmp1 + p * cmp2
        grad_log_2_t = math.log(2) * grad_s
        grad_x = cmp0 * 1.0
        return grad_output * grad_x, grad_output * grad_log_2_t, None


qunsigned = qUnsigned.apply


"""
Provide quantilized form of torch.nn.modules.conv
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math

from .number import qsigned


class Conv2d(nn.Conv2d):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 dilation=1,
                 groups=1,
                 bias=True,
                 padding_mode='zeros',
                 weight_bit_width=8,
                 bias_bit_width=16,
                 inter_bit_width=32,
                 retrain=True,
                 quant=False):
        super().__init__(in_channels=in_channels,
                         out_channels=out_channels,
                         kernel_size=kernel_size,
                         stride=stride,
                         padding=padding,
                         dilation=dilation,
                         groups=groups,
                         bias=bias,
                         padding_mode=padding_mode)
        self.weight_bit_width = weight_bit_width
        self.bias_bit_width = bias_bit_width
        self.inter_bit_width = inter_bit_width
        self.retrain = retrain
        self.quant = quant
        if retrain is True:
            self.weight_log2_t = nn.Parameter(torch.Tensor(1))
            if self.bias is not None:
                self.bias_log2_t = nn.Parameter(torch.Tensor(1))
        else:
            self.weight_log2_t = torch.Tensor(1)
            if self.bias is not None:
                self.bias_log2_t = torch.Tensor(1)

    def static(self):
        self.retrain = False
        if isinstance(self.bias_log2_t, nn.Parameter):
            self.bias_log2_t.requires_grad_(False)
        if isinstance(self.weight_log2_t, nn.Parameter):
            self.weight_log2_t.requires_grad_(False)

    def quantilize(self):
        self.quant = True
        self.weight_log2_t.requires_grad = False
        if self.bias is not None:
            self.bias_log2_t.requires_grad = False

    def floatilize(self):
        self.quant = False
        self.weight_log2_t.requires_grad = True
        if self.bias is not None:
            self.bias_log2_t.requires_grad = True

    def quant_answer(self):
        self.weight.data = qsigned(
            self.weight, self.weight_log2_t,
            self.weight_bit_width)**(self.weight_bit_width - 1 -
                                     torch.ceil(self.weight_log2_t)).int()
        self.bias.data = qsigned(
            self.bias, self.bias_log2_t,
            self.bias_bit_width)**(self.bias_bit_width - 1 -
                                   torch.ceil(self.bias_log2_t)).int()

    def conv_forward(self, input):
        weight = qsigned(self.weight, self.weight_log2_t,
                         self.weight_bit_width)
        input_log2_t = input.abs().max().log2()
        inter = qsigned(
            F.conv2d(input, weight, None, self.stride, self.padding,
                     self.dilation, self.groups),
            self.weight_log2_t + input_log2_t + math.log2(self.weight.numel()),
            self.inter_bit_width)
        if self.bias is not None:
            inter += qsigned(
                self.bias, self.bias_log2_t,
                self.bias_bit_width).unsqueeze(1).unsqueeze(2).unsqueeze(0)
        return inter

    def conv_forward_unquant(self, input):
        return F.conv2d(input, self.weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)

    def forward(self, input):
        return self.conv_forward(
            input) if self.quant else self.conv_forward_unquant(input)


if __name__ == '__main__':
    conv = Conv2d(3, 6, 3)
    x = torch.rand(4, 3, 5, 5)
    print(conv(x).shape)



import torch
import torch.nn as nn
from ..function import qsigned, qunsigned
from .foldmodule import _FoldModule


class Conv2dBNReLU(_FoldModule):
    def __init__(self, conv, bn, relu):
        super().__init__()
        if isinstance(conv, nn.Conv2d) and isinstance(
                bn, nn.BatchNorm2d) and (isinstance(
                    relu, nn.ReLU), isinstance(relu, nn.ReLU6)):
            self.conv = conv
            self.bn = bn
            self.relu = relu
            self.weight_bit_width = self.conv.weight_bit_width
            self.weight_log2_t = self.conv.weight_log2_t
            self.bias_bit_width = self.bn.bias_bit_width
            self.bias_log2_t = self.bn.bias_log2_t
            self.acti_bit_width = self.relu.acti_bit_width
            self.acti_log2_t = self.relu.acti_log2_t
        else:
            raise Exception('The folded function does not meet type check')
        self.bn_freezing = False
        self.quant = False

    def bn_freeze(self, mode=True):
        self.bn_freezing = mode

    def quantilize(self):
        self.quant = True
        self.weight_log2_t.requires_grad = True
        self.bias_log2_t.requires_grad = True

    def floatilize(self):
        self.quant = False
        self.weight_log2_t.requires_grad = False
        self.bias_log2_t.requires_grad = False

    def forward(self, input):
        if self.bn_freezing:
            bn_var = self.bn.running_var.detach().clone().data.reshape(
                -1, 1, 1, 1)
            bn_mean = self.bn.running_mean.detach().clone().data.reshape(
                -1, 1, 1, 1)
            bn_weight = self.bn.weight.detach().clone().data.reshape(
                -1, 1, 1, 1)
            bn_bias = self.bn.bias.detach().clone().data.reshape(-1, 1, 1, 1)
        else:
            bn_var = self.bn.running_var.reshape(-1, 1, 1, 1)
            bn_mean = self.bn.running_mean.reshape(-1, 1, 1, 1)
            bn_weight = self.bn.weight.reshape(-1, 1, 1, 1)
            bn_bias = self.bn.bias.reshape(-1, 1, 1, 1)
        if self.conv.bias is not None:
            conv_bias = self.conv.bias.reshape(-1, 1, 1, 1)
        else:
            conv_bias = 0.

        conv_weight = self.conv.weight * bn_weight / (bn_var +
                                                      self.bn.eps).sqrt()
        conv_bias = bn_weight * (conv_bias - bn_mean) / (
            bn_var + self.bn.eps).sqrt() + bn_bias

        if self.quant:
            conv_weight = qsigned(conv_weight, self.weight_log2_t,
                                  self.weight_bit_width)
            conv_bias = qsigned(conv_bias, self.bias_log2_t,
                                self.bias_bit_width).reshape(-1)

        inter = nn.functional.conv2d(input, conv_weight, conv_bias,
                                     self.conv.stride, self.conv.padding,
                                     self.conv.dilation, self.conv.groups)
        inter = self.relu(inter)

        if self.quant:
            output = qunsigned(inter, self.acti_log2_t, self.acti_bit_width)

        return output



import torch
import torch.nn as nn
from collections import OrderedDict
from .convbnact import Conv2dBNReLU


def fold_op(conv, bn, relu):
    folded = Conv2dBNReLU(conv, bn, relu)
    conv = folded
    bn = nn.Identity()
    relu = nn.Identity()
    return conv, bn, relu


def fold_the_network(net):
    mdic = OrderedDict(net.named_children())
    key = list(mdic.keys())
    keylen = len(key)
    flag = 0
    while flag < keylen:
        if list(OrderedDict(mdic[key[flag]].named_children()).keys()) != []:
            fold_the_network(mdic[key[flag]])
        else:
            if isinstance(mdic[key[flag]], nn.Conv2d) and flag + 1 < keylen:
                if isinstance(mdic[key[flag + 1]],
                              nn.BatchNorm2d) and flag + 2 < keylen:
                    if isinstance(mdic[key[flag + 2]], nn.ReLU) or isinstance(
                            mdic[key + 2], nn.ReLU6):
                        mdic[key[flag]], mdic[key[flag + 1]], mdic[key[
                            flag + 2]] = fold_op(mdic[key[flag]],
                                                 mdic[key[flag + 1]],
                                                 mdic[key[flag + 2]])
                        print(flag)
                        flag += 2
        flag += 1



"""
Provide quantilized form of torch.nn.modules.linear
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math

from .number import qsigned


class Linear(nn.Linear):
    def __init__(self,
                 in_features,
                 out_features,
                 bias=True,
                 weight_bit_width=8,
                 bias_bit_width=16,
                 inter_bit_width=32,
                 retrain=True,
                 quant=False):
        super().__init__(in_features, out_features, bias=bias)
        self.weight_bit_width = weight_bit_width
        self.bias_bit_width = bias_bit_width
        self.inter_bit_width = inter_bit_width
        self.retrain = retrain
        self.quant = quant
        if retrain is True:
            self.weight_log2_t = nn.Parameter(torch.Tensor(1))
            if self.bias is not None:
                self.bias_log2_t = nn.Parameter(torch.Tensor(1))
        else:
            self.weight_log2_t = torch.Tensor(1)
            if self.bias is not None:
                self.bias_log2_t = torch.Tensor(1)

    def static(self):
        self.retrain = False
        if isinstance(self.bias_log2_t, nn.Parameter):
            self.bias_log2_t.requires_grad_(False)
        if isinstance(self.weight_log2_t, nn.Parameter):
            self.weight_log2_t.requires_grad_(False)

    def quantilize(self):
        self.quant = True
        self.weight_log2_t.requires_grad = True
        if self.bias is not None:
            self.bias_log2_t.requires_grad = True

    def floatilize(self):
        self.quant = False
        self.weight_log2_t.requires_grad = False
        if self.bias is not None:
            self.bias_log2_t.requires_grad = False

    def quant_answer(self):
        self.weight.data = qsigned(
            self.weight, self.weight_log2_t,
            self.weight_bit_width)**(self.weight_bit_width - 1 -
                                     torch.ceil(self.weight_log2_t)).int()
        self.bias.data = qsigned(
            self.bias, self.bias_log2_t,
            self.bias_bit_width)**(self.bias_bit_width - 1 -
                                   torch.ceil(self.bias_log2_t)).int()

    def linear_forward(self, input):
        input_log2_t = input.abs().max().log2()
        weight = qsigned(self.weight, self.weight_log2_t,
                         self.weight_bit_width)
        inter = qsigned(
            F.linear(input, weight, None),
            self.weight_log2_t + input_log2_t + math.log2(self.weight.numel()),
            self.inter_bit_width)
        if self.bias is not None:
            inter += qsigned(self.bias, self.bias_log2_t,
                             self.bias_bit_width).unsqueeze(0)
        return inter

    def linear_forward_unquant(self, input):
        return F.linear(input, self.weight, self.bias)

    def forward(self, input):
        return self.linear_forward(
            input) if self.quant else self.linear_forward_unquant(input)


if __name__ == '__main__':
    lin = Linear(3, 5)
    x = torch.rand(3, 3)
    print(lin(x).shape)



#!/usr/bin/env python3

import os
from setuptools import setup, find_packages

# get key package details from tqt/__version__.py
about = {}  # type: ignore
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'tqt', '__version__.py')) as f:
    exec(f.read(), about)

# load the README file and use it as the long_description for PyPI
with open('README.md', 'r') as f:
    readme = f.read()

# package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
setup(name=about['__title__'],
      description=about['__description__'],
      long_description=readme,
      long_description_content_type='text/markdown',
      version=about['__version__'],
      author=about['__author__'],
      author_email=about['__author_email__'],
      url=about['__url__'],
      packages=find_packages(where='.', exclude=(), include=('*', )),
      include_package_data=True,
      python_requires=">=3.7.*",
      install_requires=['torch'],
      license=about['__license__'],
      zip_safe=False,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: GPU',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
      ],
      keywords='quantization NN torch')
